import os
from glob import glob
import pandas as pd
import click

input = 'output/all_annotations/*/annotations.tsv'
output = 'results/all_annotations.tsv'
@click.command()
@click.option('--input',
              help='A glob/regex string for input.')
@click.option('--output', default='all_annotations.tsv',
              help='A name for the output, like all_annotations.tsv')
def append_annotations_lowmem(input:str, output:str):
    paths = glob(input)
    all_headers = pd.concat([pd.read_csv(i, nrows=0, sep='\t', index_col=0) for i in paths])
    all_headers.to_csv(output, sep='\t')
    for i in paths:
        pd.concat([
            all_headers,
            pd.read_csv(i, sep='\t', index_col=0)
        ]).to_csv(output, mode='a', header=False, sep='\t')

if __name__ == '__main__':
    append_annotations_lowmem()



"""A trimed down distillation scripts for use with CAMPERS"""
import os
from os import path, mkdir
import click
import pandas as pd
from collections import Counter, defaultdict
import altair as alt
import networkx as nx
from itertools import tee
import re
import numpy as np
from datetime import datetime
from camper_dramkit.camper_annotate import CAMPER_NAME


# TODO: add RBH information to output
# TODO: add flag to output table and not xlsx
# TODO: add flag to output heatmap table

FRAME_COLUMNS = ['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader']
RRNA_TYPES = ['5S rRNA', '16S rRNA', '23S rRNA']
HEATMAP_MODULES = ['M00001', 'M00004', 'M00008', 'M00009', 'M00012', 'M00165', 'M00173', 'M00374', 'M00375',
                   'M00376', 'M00377', 'M00422', 'M00567']
HEATMAP_CELL_HEIGHT = 10
HEATMAP_CELL_WIDTH = 10
KO_REGEX = r'^K\d\d\d\d\d$'
ETC_COVERAGE_COLUMNS = ['module_id', 'module_name', 'complex', 'genome', 'path_length', 'path_length_coverage',
                        'percent_coverage', 'genes', 'missing_genes', 'complex_module_name']
TAXONOMY_LEVELS = ['d', 'p', 'c', 'o', 'f', 'g', 's']

DEFAULT_CAMPER_DIST = os.path.join(os.path.dirname(__file__) , "data", "CAMPER_distillate.tsv")


# UTILS
def get_ordered_uniques(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def get_ids_from_row(row):
    id_list = list()
    # get kegg gene ids
    if 'kegg_genes_id' in row and not pd.isna(row['kegg_genes_id']):
        id_list += row['kegg_genes_id']
    # get kegg orthology ids
    if 'ko_id' in row and not pd.isna(row['ko_id']):
        id_list += [j for j in row['ko_id'].split(',')]
    # get ec numbers
    if 'kegg_hit' in row and not pd.isna(row['kegg_hit']):
        id_list += [i[1:-1] for i in re.findall(r'\[EC:\d*.\d*.\d*.\d*\]', row['kegg_hit'])]
    # get merops ids
    if 'peptidase_family' in row and not pd.isna(row['peptidase_family']):
        id_list += [j for j in row['peptidase_family'].split(';')]
    # get cazy ids
    if 'cazy_hits' in row and not pd.isna(row['cazy_hits']):
        id_list += [i[1:-1].split('_')[0] for i in re.findall(r'\[[A-Z]*\d*?\]', row['cazy_hits'])]
    # get pfam ids
    if 'pfam_hits' in row and not pd.isna(row['pfam_hits']):
        id_list += [j[1:-1].split('.')[0]
                    for j in re.findall(r'\[PF\d\d\d\d\d.\d*\]', row['pfam_hits'])]
    # custom campers id
    if f"{CAMPER_NAME}_id" in row:
        id_list += [row[f"{CAMPER_NAME}_id"]]
    return set(id_list)


def get_ids_from_annotation(frame):
    id_list = list()
    # get kegg gene ids
    if 'kegg_genes_id' in frame:
        id_list += [j.strip() for i in frame.kegg_genes_id.dropna() for j in i.split(',')]
    # get kegg orthology ids
    if 'ko_id' in frame:
        id_list += [j.strip() for i in frame.ko_id.dropna() for j in i.split(',')]
    # get kegg ec numbers
    if 'kegg_hit' in frame:
        for kegg_hit in frame.kegg_hit.dropna():
            id_list += [i[1:-1] for i in re.findall(r'\[EC:\d*.\d*.\d*.\d*\]', kegg_hit)]
    # get merops ids
    if 'peptidase_family' in frame:
        id_list += [j.strip() for i in frame.peptidase_family.dropna() for j in i.split(';')]
    # get cazy ids
    if 'cazy_id' in frame:
        id_list += [j for i in frame.cazy_id.dropna() for j in i.split('; ')]
    # get cazy ec numbers
    if 'cazy_hits' in frame:
        id_list += [f"{j[1:3]}:{j[4:-1]}" for i in frame.cazy_hits.dropna()
                    for j in re.findall(r'\(EC [\d+\.]+[\d-]\)', i)]
    # get pfam ids
    if 'pfam_hits' in frame:
        id_list += [j[1:-1].split('.')[0] for i in frame.pfam_hits.dropna()
                    for j in re.findall(r'\[PF\d\d\d\d\d.\d*\]', i)]
    if f"{CAMPER_NAME}_id" in frame:
        id_list += frame[f"{CAMPER_NAME}_id"].dropna().apply(lambda x: x.strip()).tolist()
    return Counter(id_list)


def fill_genome_summary_frame(annotations, genome_summary_frame, groupby_column):
    genome_summary_id_sets = [set([k.strip() for k in j.split(',')]) for j in genome_summary_frame['gene_id']]
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        id_dict = get_ids_from_annotation(frame)
        counts = list()
        for i in genome_summary_id_sets:
            identifier_count = 0
            for j in i:
                if j in id_dict:
                    identifier_count += id_dict[j]
            counts.append(identifier_count)
        genome_summary_frame[genome] = counts
    return genome_summary_frame


def fill_genome_summary_frame_gene_names(annotations, genome_summary_frame, groupby_column):
    genome_summary_id_sets = [set([k.strip() for k in j.split(',')]) for j in genome_summary_frame['gene_id']]
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        # make dict of identifiers to gene names
        id_gene_dict = defaultdict(list)
        for gene, row in frame.iterrows():
            ids = get_ids_from_row(row)
            for id_ in ids:
                id_gene_dict[id_].append(gene)
        # fill in genome summary_frame
        values = list()
        for id_set in genome_summary_id_sets:
            this_value = list()
            for id_ in id_set:
                this_value += id_gene_dict[id_]
            values.append(','.join(this_value))
        genome_summary_frame[genome] = values
    return genome_summary_frame


def make_genome_summary(annotations, genome_summary_frame, groupby_column='fasta'):
    summary_frames = list()
    # get ko summaries
    summary_frames.append(fill_genome_summary_frame(annotations, genome_summary_frame.copy(), groupby_column))

    # merge summary frames
    summarized_genomes = pd.concat(summary_frames, sort=False)
    return summarized_genomes


def write_summarized_genomes_to_xlsx(summarized_genomes, output_file):
    # turn all this into an xlsx
    with pd.ExcelWriter(output_file) as writer:
        for sheet, frame in summarized_genomes.groupby('sheet', sort=False):
            frame = frame.sort_values(['header', 'subheader', 'module', 'gene_id'])
            frame = frame.drop(['sheet'], axis=1)
            frame.to_excel(writer, sheet_name=sheet, index=False)


# TODO: add assembly stats like N50, longest contig, total assembled length etc
def make_genome_stats(annotations, rrna_frame=None, trna_frame=None, groupby_column='fasta'):
    rows = list()
    columns = ['genome']
    if 'scaffold' in annotations.columns:
        columns.append('number of scaffolds')
    if 'bin_taxonomy' in annotations.columns:
        columns.append('taxonomy')
    if 'bin_completeness' in annotations.columns:
        columns.append('completeness score')
    if 'bin_contamination' in annotations.columns:
        columns.append('contamination score')
    if rrna_frame is not None:
        columns += RRNA_TYPES
    if trna_frame is not None:
        columns.append('tRNA count')
    if 'bin_completeness' in annotations.columns and 'bin_contamination' in annotations.columns \
       and rrna_frame is not None and trna_frame is not None:
        columns.append('assembly quality')
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        row = [genome]
        if 'scaffold' in frame.columns:
            row.append(len(set(frame['scaffold'])))
        if 'bin_taxonomy' in frame.columns:
            row.append(frame['bin_taxonomy'][0])
        if 'bin_completeness' in frame.columns:
            row.append(frame['bin_completeness'][0])
        if 'bin_contamination' in frame.columns:
            row.append(frame['bin_contamination'][0])
        has_rrna = list()
        if rrna_frame is not None:
            genome_rrnas = rrna_frame.loc[rrna_frame.fasta == genome]
            for rrna in RRNA_TYPES:
                sixteens = genome_rrnas.loc[genome_rrnas.type == rrna]
                if sixteens.shape[0] == 0:
                    row.append('')
                    has_rrna.append(False)
                elif sixteens.shape[0] == 1:
                    row.append('%s (%s, %s)' % (sixteens['scaffold'].iloc[0], sixteens.begin.iloc[0],
                                                sixteens.end.iloc[0]))
                    has_rrna.append(True)
                else:
                    row.append('%s present' % sixteens.shape[0])
                    has_rrna.append(False)
        if trna_frame is not None:
            row.append(trna_frame.loc[trna_frame[groupby_column] == genome].shape[0])  # TODO: remove psuedo from count?
        if 'assembly quality' in columns:
            if frame['bin_completeness'][0] > 90 and frame['bin_contamination'][0] < 5 and np.all(has_rrna) and \
               len(set(trna_frame.loc[trna_frame[groupby_column] == genome].Type)) >= 18:
                row.append('high')
            elif frame['bin_completeness'][0] >= 50 and frame['bin_contamination'][0] < 10:
                row.append('med')
            else:
                row.append('low')
        rows.append(row)
    genome_stats = pd.DataFrame(rows, columns=columns)
    return genome_stats


def build_module_net(module_df):
    """Starts with a data from including a single module"""
    # build net from a set of module paths
    num_steps = max([int(i.split(',')[0]) for i in set(module_df['path'])])
    module_net = nx.DiGraph(num_steps=num_steps, module_id=list(module_df['module'])[0],
                            module_name=list(module_df['module_name'])[0])
    # go through all path/step combinations
    for module_path, frame in module_df.groupby('path'):
        split_path = [int(i) for i in module_path.split(',')]
        step = split_path[0]
        module_net.add_node(module_path, kos=set(frame['ko']))
        # add incoming edge
        if step != 0:
            module_net.add_edge('end_step_%s' % (step - 1), module_path)
        # add outgoing edge
        module_net.add_edge(module_path, 'end_step_%s' % step)
    return module_net


def get_module_step_coverage(kos, module_net):
    # prune network based on what kos were observed
    pruned_module_net = module_net.copy()
    module_kos_present = set()
    for node, data in module_net.nodes.items():
        if 'kos' in data:
            ko_overlap = data['kos'] & kos
            if len(ko_overlap) == 0:
                pruned_module_net.remove_node(node)
            else:
                module_kos_present = module_kos_present | ko_overlap
    # count number of missing steps
    missing_steps = 0
    for node, data in pruned_module_net.nodes.items():
        if ('end_step' in node) and (pruned_module_net.in_degree(node) == 0):
            missing_steps += 1
    # get statistics
    num_steps = pruned_module_net.graph['num_steps'] + 1
    num_steps_present = num_steps - missing_steps
    coverage = num_steps_present / num_steps
    return num_steps, num_steps_present, coverage, sorted(module_kos_present)


def make_module_coverage_df(annotation_df, module_nets):
    kos_to_genes = defaultdict(list)
    for gene_id, ko_list in annotation_df['ko_id'].iteritems():
        if type(ko_list) is str:
            for ko in ko_list.split(','):
                kos_to_genes[ko].append(gene_id)
    coverage_dict = {}
    for i, (module, net) in enumerate(module_nets.items()):
        module_steps, module_steps_present, module_coverage, module_kos = \
            get_module_step_coverage(set(kos_to_genes.keys()), net)
        module_genes = sorted([gene for ko in module_kos for gene in kos_to_genes[ko]])
        coverage_dict[module] = [net.graph['module_name'], module_steps, module_steps_present, module_coverage,
                                 len(module_kos), ','.join(module_kos), ','.join(module_genes)]
    coverage_df = pd.DataFrame.from_dict(coverage_dict, orient='index',
                                         columns=['module_name', 'steps', 'steps_present', 'step_coverage', 'ko_count',
                                                  'kos_present', 'genes_present'])
    return coverage_df


def make_module_coverage_frame(annotations, module_nets, groupby_column='fasta'):
    # go through each scaffold to check for modules
    module_coverage_dict = dict()
    for group, frame in annotations.groupby(groupby_column, sort=False):
        module_coverage_dict[group] = make_module_coverage_df(frame, module_nets)
    module_coverage = pd.concat(module_coverage_dict)
    module_coverage.index = module_coverage.index.set_names(['genome', 'module'])
    return module_coverage.reset_index()


def make_module_coverage_heatmap(module_coverage, mag_order=None):
    num_mags_in_frame = len(set(module_coverage['genome']))
    c = alt.Chart(module_coverage, title='Module').encode(
        x=alt.X('module_name', title=None, sort=None, axis=alt.Axis(labelLimit=0, labelAngle=90)),
        y=alt.Y('genome', title=None, sort=mag_order, axis=alt.Axis(labelLimit=0)),
        tooltip=[alt.Tooltip('genome', title='Genome'),
                 alt.Tooltip('module_name', title='Module Name'),
                 alt.Tooltip('steps', title='Module steps'),
                 alt.Tooltip('steps_present', title='Steps present')
                 ]
    ).mark_rect().encode(color=alt.Color('step_coverage', legend=alt.Legend(title='% Complete'),
                                         scale=alt.Scale(domain=(0, 1)))).properties(
        width=HEATMAP_CELL_WIDTH * len(HEATMAP_MODULES),
        height=HEATMAP_CELL_HEIGHT * num_mags_in_frame)
    return c


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def first_open_paren_is_all(str_):
    """Go through string and return true"""
    curr_level = 1
    for char in str_[1:-1]:
        if char == ')':
            curr_level -= 1
        elif char == '(':
            curr_level += 1
        if curr_level == 0:
            return False
    return True


def split_into_steps(definition, split_char=' '):
    """Very fancy split on string of chars"""
    curr_level = 0
    step_starts = [-1]
    for i, char in enumerate(definition):
        if char == '(':
            curr_level += 1
        if char == ')':
            curr_level -= 1
        if (curr_level == 0) and (char in split_char):
            step_starts.append(i)
    step_starts.append(len(definition))
    steps = list()
    for a, b in pairwise(step_starts):
        step = definition[a+1:b]
        if step.startswith('(') and step.endswith(')'):
            if first_open_paren_is_all(step):
                step = step[1:-1]
        steps.append(step)
    return steps


def is_ko(ko):
    return re.match(KO_REGEX, ko) is not None


def make_module_network(definition, network: nx.DiGraph = None, parent_nodes=('start',)):
    # TODO: Figure out how to add 'end' step to last step at end
    if network is None:
        network = nx.DiGraph()
    last_steps = []
    for step in split_into_steps(definition, ','):
        prev_steps = parent_nodes
        for substep in split_into_steps(step, '+'):
            if is_ko(substep):
                for prev_step in prev_steps:
                    network.add_edge(prev_step, substep)
                prev_steps = [substep]
            else:
                network, prev_steps = make_module_network(substep, network, prev_steps)
        last_steps += prev_steps
    return network, last_steps


def get_module_coverage(module_net: nx.DiGraph, genes_present: set):
    max_coverage = -1
    max_coverage_genes = list()
    max_coverage_missing_genes = list()
    max_path_len = 0
    for net_path in nx.all_simple_paths(module_net, source='start', target='end'):
        net_path = set(net_path[1:-1])
        overlap = net_path & genes_present
        coverage = len(overlap) / len(net_path)
        if coverage > max_coverage:
            max_coverage = coverage
            max_coverage_genes = overlap
            max_coverage_missing_genes = net_path - genes_present
            max_path_len = len(net_path)
    return max_path_len, len(max_coverage_genes), max_coverage, max_coverage_genes, max_coverage_missing_genes


def make_etc_coverage_df(etc_module_df, annotations, groupby_column='fasta'):
    etc_coverage_df_rows = list()
    for _, module_row in etc_module_df.iterrows():
        definition = module_row['definition']
        # remove optional subunits
        definition = re.sub(r'-K\d\d\d\d\d', '', definition)
        module_net, _ = make_module_network(definition)
        # add end node
        no_out = [node for node in module_net.nodes() if module_net.out_degree(node) == 0]
        for node in no_out:
            module_net.add_edge(node, 'end')
        # go through each genome and check pathway coverage
        for group, frame in annotations.groupby(groupby_column):
            # get annotation genes
            grouped_ids = set(get_ids_from_annotation(frame).keys())
            path_len, path_coverage_count, path_coverage_percent, genes, missing_genes = \
                get_module_coverage(module_net, grouped_ids)
            complex_module_name = 'Complex %s: %s' % (module_row['complex'].replace('Complex ', ''),
                                                      module_row['module_name'])
            etc_coverage_df_rows.append([module_row['module_id'], module_row['module_name'],
                                         module_row['complex'].replace('Complex ', ''), group, path_len,
                                         path_coverage_count, path_coverage_percent, ','.join(sorted(genes)),
                                         ','.join(sorted(missing_genes)), complex_module_name])
    return pd.DataFrame(etc_coverage_df_rows, columns=ETC_COVERAGE_COLUMNS)


def make_etc_coverage_heatmap(etc_coverage, mag_order=None, module_order=None):
    num_mags_in_frame = len(set(etc_coverage['genome']))
    charts = list()
    for i, (etc_complex, frame) in enumerate(etc_coverage.groupby('complex')):
        # if this is the first chart then make y-ticks otherwise none
        c = alt.Chart(frame, title=etc_complex).encode(
            x=alt.X('module_name', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90),
                    sort=module_order),
            y=alt.Y('genome', axis=alt.Axis(title=None, labels=False, ticks=False), sort=mag_order),
            tooltip=[alt.Tooltip('genome', title='Genome'),
                     alt.Tooltip('module_name', title='Module Name'),
                     alt.Tooltip('path_length', title='Module Subunits'),
                     alt.Tooltip('path_length_coverage', title='Subunits present'),
                     alt.Tooltip('genes', title='Genes present'),
                     alt.Tooltip('missing_genes', title='Genes missing')
                     ]
        ).mark_rect().encode(color=alt.Color('percent_coverage', legend=alt.Legend(title='% Complete'),
                                             scale=alt.Scale(domain=(0, 1)))).properties(
            width=HEATMAP_CELL_WIDTH * len(set(frame['module_name'])),
            height=HEATMAP_CELL_HEIGHT * num_mags_in_frame)
        charts.append(c)
    concat_title = alt.TitleParams('ETC Complexes', anchor='middle')
    return alt.hconcat(*charts, spacing=5, title=concat_title)



def make_functional_df(annotations, function_heatmap_form, groupby_column='fasta'):
    # clean up function heatmap form
    function_heatmap_form = function_heatmap_form.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
    function_heatmap_form = function_heatmap_form.fillna('')
    # build dict of ids per genome
    genome_to_id_dict = dict()
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        id_list = get_ids_from_annotation(frame).keys()
        genome_to_id_dict[genome] = set(id_list)
    # build long from data frame
    rows = list()
    for function, frame in function_heatmap_form.groupby('function_name', sort=False):
        for bin_name, id_set in genome_to_id_dict.items():
            presents_in_bin = list()
            functions_present = set()
            for _, row in frame.iterrows():
                function_id_set = set([i.strip() for i in row.function_ids.strip().split(',')])
                present_in_bin = id_set & function_id_set
                functions_present = functions_present | present_in_bin
                presents_in_bin.append(len(present_in_bin) > 0)
            function_in_bin = np.all(presents_in_bin)
            row = frame.iloc[0]
            rows.append([row.category, row.subcategory, row.function_name, ', '.join(functions_present),
                         '; '.join(get_ordered_uniques(frame.long_function_name)),
                         '; '.join(get_ordered_uniques(frame.gene_symbol)), bin_name, function_in_bin,
                         '%s: %s' % (row.category, row.function_name)])
    return pd.DataFrame(rows, columns=list(function_heatmap_form.columns) + ['genome', 'present',
                                                                             'category_function_name'])


def make_functional_heatmap(functional_df, mag_order=None):
    # build heatmaps
    charts = list()
    for i, (group, frame) in enumerate(functional_df.groupby('category', sort=False)):
        # set variables for chart
        function_order = get_ordered_uniques(list(frame.function_name))
        num_mags_in_frame = len(set(frame['genome']))
        chart_width = HEATMAP_CELL_WIDTH * len(function_order)
        chart_height = HEATMAP_CELL_HEIGHT * num_mags_in_frame
        # if this is the first chart then make y-ticks otherwise none
        if i == 0:
            y = alt.Y('genome', title=None, sort=mag_order,
                      axis=alt.Axis(labelLimit=0, labelExpr="replace(datum.label, /_\d*$/gi, '')"))
        else:
            y = alt.Y('genome', axis=alt.Axis(title=None, labels=False, ticks=False), sort=mag_order)
        # set up colors for chart
        rect_colors = alt.Color('present',
                                legend=alt.Legend(title="Function is Present", symbolType='square',
                                                  values=[True, False]),
                                sort=[True, False],
                                scale=alt.Scale(range=['#2ca25f', '#e5f5f9']))
        # define chart
        # TODO: Figure out how to angle title to take up less space
        c = alt.Chart(frame, title=alt.TitleParams(group)).encode(
            x=alt.X('function_name', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90), sort=function_order),
            tooltip=[alt.Tooltip('genome', title='Genome'),
                     alt.Tooltip('category', title='Category'),
                     alt.Tooltip('subcategory', title='Subcategory'),
                     alt.Tooltip('function_ids', title='Function IDs'),
                     alt.Tooltip('function_name', title='Function'),
                     alt.Tooltip('long_function_name', title='Description'),
                     alt.Tooltip('gene_symbol', title='Gene Symbol')]
        ).mark_rect().encode(y=y, color=rect_colors).properties(
            width=chart_width,
            height=chart_height)
        charts.append(c)
    # merge and return
    function_heatmap = alt.hconcat(*charts, spacing=5)
    return function_heatmap


# TODO: refactor this to handle splitting large numbers of genomes into multiple heatmaps here
def fill_liquor_dfs(annotations, module_nets, etc_module_df, function_heatmap_form, groupby_column='fasta'):
    module_coverage_frame = make_module_coverage_frame(annotations, module_nets, groupby_column)

    # make ETC frame
    etc_coverage_df = make_etc_coverage_df(etc_module_df, annotations, groupby_column)

    # make functional frame
    function_df = make_functional_df(annotations, function_heatmap_form, groupby_column)

    return module_coverage_frame, etc_coverage_df, function_df


def rename_genomes_to_taxa(function_df, labels, mag_order):
    function_df = function_df.copy()
    new_genome_column = [labels[i] for i in function_df['genome']]
    function_df['genome'] = pd.Series(new_genome_column, index=function_df.index)
    mag_order = [labels[i] for i in mag_order]
    return function_df, mag_order


def make_liquor_heatmap(module_coverage_frame, etc_coverage_df, function_df, mag_order=None, labels=None):
    module_coverage_heatmap = make_module_coverage_heatmap(module_coverage_frame, mag_order)
    etc_heatmap = make_etc_coverage_heatmap(etc_coverage_df, mag_order=mag_order)
    if labels is not None:
        function_df, mag_order = rename_genomes_to_taxa(function_df, labels, mag_order)
    function_heatmap = make_functional_heatmap(function_df, mag_order)

    liquor = alt.hconcat(alt.hconcat(module_coverage_heatmap, etc_heatmap), function_heatmap)
    return liquor


def make_liquor_df(module_coverage_frame, etc_coverage_df, function_df):
    liquor_df = pd.concat([module_coverage_frame.pivot(index='genome', columns='module_name', values='step_coverage'),
                           etc_coverage_df.pivot(index='genome', columns='complex_module_name',
                                                 values='percent_coverage'),
                           function_df.pivot(index='genome', columns='category_function_name', values='present')],
                          axis=1, sort=False)
    return liquor_df


def get_phylum_and_most_specific(taxa_str):
    taxa_ranks = [i[3:] for i in taxa_str.split(';')]
    phylum = taxa_ranks[1]
    most_specific_rank = TAXONOMY_LEVELS[sum([len(i) > 0 for i in taxa_ranks])-1]
    most_specific_taxa = taxa_ranks[sum([len(i) > 0 for i in taxa_ranks]) - 1]
    if most_specific_rank == 'd':
        return 'd__%s;p__' % most_specific_taxa
    if most_specific_rank == 'p':
        return 'p__%s;c__' % most_specific_taxa
    else:
        return 'p__%s;%s__%s' % (phylum, most_specific_rank, most_specific_taxa)


def make_strings_no_repeats(genome_taxa_dict):
    labels = dict()
    seen = Counter()
    for genome, taxa_string in genome_taxa_dict.items():
        final_taxa_string = '%s_%s' % (taxa_string, str(seen[taxa_string]))
        seen[taxa_string] += 1
        labels[genome] = final_taxa_string
    return labels


def summarize_genomes(annotations,  output_tsv, groupby_column='fasta',
                      camper_distillate=DEFAULT_CAMPER_DIST):
    start_time = datetime.now()

    # read in data
    annotations = pd.read_csv(annotations, sep='\t', index_col=0)
    if 'bin_taxnomy' in annotations:
        annotations = annotations.sort_values('bin_taxonomy')

    genome_summary_form = pd.read_csv(camper_distillate, sep='\t')
    summarized_genomes = make_genome_summary(annotations, genome_summary_form, groupby_column)
    summarized_genomes.drop(['sheet'], axis=1, inplace=True)
    summarized_genomes.to_csv(output_tsv, sep='\t', index=False)



@click.command()
@click.option('-a', '--annotations', help="annotations file, from dram annotate", 
              required=True)
@click.option('-o', '--output_tsv', help="output tsv loctation, it can overwright!", required=True)
@click.option('--groupby_column', type=str, default='fasta',
              help="past_annotations to append new annotations to.")
@click.option('--camper_distillate', default=DEFAULT_CAMPER_DIST,
              help="CAMPER Distillate file, if not specified the default will be used")
def summarize_genomes_cmd(annotations,  output_tsv, groupby_column='fasta',
                      camper_distillate=DEFAULT_CAMPER_DIST):
    summarize_genomes(annotations=annotations,  
                      output_tsv=output_tsv, 
                      groupby_column=groupby_column,
                      camper_distillate=camper_distillate)

if __name__ == "__main__":
    summarize_genomes_cmd()

"""
import os
os.system("python workflow/scripts/camper_distill.py -a output/t1/annotations.tsv -o results/CAMPER_distillate.tsv --custom_distillate CAMPER/test_CAMPER_distillate.tsv")
os.system("rm -rf output/t1/dist")


"""



from setuptools import setup, find_packages
from camper_dramkit import __version__ as version
import os

__author__ = 'rmflynn'
__version__ = version


here = os.path.abspath(os.path.dirname(__file__))
data_folder = os.path.join(here, 'camper_dramkit', 'data')
if not os.path.exists(data_folder):
    os.mkdir(data_folder)
    os.rename(os.path.join(here, '..', 'CAMPER_blast.faa'),
              os.path.join(data_folder, 'CAMPER_blast.faa'))
    os.rename(os.path.join(here, '..', 'CAMPER_blast_scores.tsv'),
              os.path.join(data_folder, 'CAMPER_blast_scores.tsv'))
    os.rename(os.path.join(here, '..', 'CAMPER.hmm'),
              os.path.join(data_folder, 'CAMPER.hmm'))
    os.rename(os.path.join(here, '..', 'CAMPER_distillate.tsv'),
              os.path.join(data_folder, 'CAMPER_distillate.tsv'))
    os.rename(os.path.join(here, '..', 'CAMPER_hmm_scores.tsv'),
              os.path.join(data_folder, 'CAMPER_hmm_scores.tsv'))

with open(os.path.join(here, "..", 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="camper_dramkit",
    version=__version__,
    description="A tool to use the CAMPER dataset",
    long_description=long_description,
    long_description_content_type='text/markdown',  # Optional (see note above)
    packages=['camper_dramkit'],
    package_dir={'camper_dramkit': 'camper_dramkit'},
    package_data={'camper_dramkit': ['data/*']},
    python_requires='>=3.8',
    install_requires=['pandas', 'altair', 'sqlalchemy', 'networkx', 'openpyxl', 'numpy', 'click'],
    entry_points={
        'console_scripts': [
            'camper_distill = camper_dramkit.camper_distill:summarize_genomes_cmd',
            'camper_annotate = camper_dramkit.camper_annotate:annotate_genes_cmd',
            'combine_annotations_lowmem = camper_dramkit.combine_annotations_lowmem:append_annotations_lowmem'
        ],
    },
    author="Rory Flynn",
    author_email='rory.flynn@colostate.edu',
    url="https://github.com/WrightonLabCSU/CAMPER",
    download_url="https://github.com/WrightonLabCSU/CAMPER/tarball/%s" % __version__
)



"""A cut down of dram annotator that only adds CAMPER Annotations"""
import re
import os
import warnings
import subprocess
from glob import glob
from functools import partial
from datetime import datetime
from os import path, mkdir, stat
from typing import Callable
from skbio.io import read as read_sequence
from shutil import rmtree
from skbio.io import write as write_sequence
import pandas as pd
import numpy as np
import click
# Remove

CAMPER_NAME = "CAMPER"
DEFAULT_CUSTOM_FA_DB_LOC = os.path.join(os.path.dirname(__file__),  "data", "CAMPER_blast.faa")
DEFAULT_CUSTOM_HMM_LOC = os.path.join(os.path.dirname(__file__), "data", "CAMPER.hmm")
DEFAULT_CUSTOM_HMM_CUTOFFS_LOC = os.path.join(os.path.dirname(__file__), "data", "CAMPER_hmm_scores.tsv")
DEFAULT_CUSTOM_FA_DB_CUTOFFS_LOC = os.path.join(os.path.dirname(__file__), "data", "CAMPER_blast_scores.tsv")

BOUTFMT6_COLUMNS = ['qId', 'tId', 'seqIdentity', 'alnLen', 'mismatchCnt',
                    'gapOpenCnt', 'qStart', 'qEnd', 'tStart', 'tEnd', 'eVal',
                    'bitScore']
HMMSCAN_ALL_COLUMNS = ['query_id', 'query_ascession', 'query_length', 
                       'target_id', 'target_ascession', 'target_length',
                       'full_evalue', 'full_score', 'full_bias', 
                       'domain_number', 'domain_count', 'domain_cevalue',
                       'domain_ievalue', 'domain_score', 'domain_bias',
                       'target_start', 'target_end', 'alignment_start',
                       'alignment_end', 'query_start', 'query_end', 'accuracy',
                       'description']
HMMSCAN_COLUMN_TYPES = [str, str, int, str, str, int, float, float, float, int,
                        int, float, float, float, float, int, int, int, int, 
                        int, int, float, str]


def remove_suffix(text, suffix):
    if text.endswith(suffix):
        return text[:-1*len(suffix)]
    return text  # or whatever


def merge_files(files_to_merge, outfile, has_header=False):
    """It's in the name, if has_header assumes all files have the same header"""
    with open(outfile, 'w') as outfile_handle:
        if has_header:
            outfile_handle.write(open(files_to_merge[0]).readline())
        for file in files_to_merge:
            with open(file) as f:
                if has_header:
                    _ = f.readline()
                outfile_handle.write(f.read())


def run_process(command, shell=False, capture_stdout=True, check=True, verbose=False):
    """Standardization of parameters for using subprocess.run, provides verbose mode and option to run via shell"""
    stdout = subprocess.DEVNULL
    stderr = subprocess.DEVNULL
    if verbose:
        stdout = None
        stderr = None
    if capture_stdout:
        return subprocess.run(command, check=check, shell=shell, stdout=subprocess.PIPE,
                              stderr=stderr).stdout.decode(errors='ignore')
    else:
        subprocess.run(command, check=check, shell=shell, stdout=stdout, stderr=stderr)


def make_mmseqs_db(fasta_loc, output_loc, create_index=True, threads=10, verbose=False):
    """Takes a fasta file and makes a mmseqs2 database for use in blast searching and hmm searching with mmseqs2"""
    run_process(['mmseqs', 'createdb', fasta_loc, output_loc], verbose=verbose)
    if create_index:
        tmp_dir = path.join(path.dirname(output_loc), 'tmp')
        run_process(['mmseqs', 'createindex', output_loc, tmp_dir, '--threads', str(threads)], verbose=verbose)


def multigrep(search_terms, search_against, split_char='\n', output='.'):
    # TODO: multiprocess this over the list of search terms
    """Search a list of exact substrings against a database, takes name of mmseqs db index with _h to search against"""
    hits_file = path.join(output, 'hits.txt')
    with open(hits_file, 'w') as f:
        f.write('%s\n' % '\n'.join(search_terms))
    results = run_process(['grep', '-a', '-F', '-f', hits_file, search_against], capture_stdout=True, verbose=False)
    processed_results = [i.strip() for i in results.strip().split(split_char)
                         if len(i) > 0]
    # remove(hits_file)
    return {i.split()[0]: i for i in processed_results if i != ''}



def filter_fasta(fasta_loc, min_len=5000, output_loc=None):
    """Removes sequences shorter than a set minimum from fasta files, outputs an object or to a file"""
    kept_seqs = (seq for seq in read_sequence(fasta_loc, format='fasta') if len(seq) >= min_len)
    if output_loc is None:
        return list(kept_seqs)
    else:
        write_sequence(kept_seqs, format='fasta', into=output_loc)


def run_prodigal(fasta_loc, output_dir, mode='meta', trans_table='11', verbose=False):
    """Runs the prodigal gene caller on a given fasta file, outputs resulting files to given directory"""
    output_gff = path.join(output_dir, 'genes.gff')
    output_fna = path.join(output_dir, 'genes.fna')
    output_faa = path.join(output_dir, 'genes.faa')

    run_process(['prodigal', '-i', fasta_loc, '-p', mode, '-g', trans_table, '-f', 'gff', '-o', output_gff, '-a',
                 output_faa, '-d', output_fna], verbose=verbose)
    return output_gff, output_fna, output_faa


def get_best_hits(query_db, target_db, output_dir='.', query_prefix='query', target_prefix='target',
                  bit_score_threshold=60, threads=10, verbose=False):
    """Uses mmseqs2 to do a blast style search of a query db against a target db, filters to only include best hits
    Returns a file location of a blast out format 6 file with search results
    """
    # make query to target db
    tmp_dir = path.join(output_dir, 'tmp')
    query_target_db = path.join(output_dir, '%s_%s.mmsdb' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'search', query_db, target_db, query_target_db, tmp_dir, '--threads', str(threads)],
                verbose=verbose)
    # filter query to target db to only best hit
    query_target_fa_db_top = path.join(output_dir, '%s_%s.tophit.mmsdb' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'filterdb', query_target_db, query_target_fa_db_top, '--extract-lines', '1'], verbose=verbose)
    # filter query to target db to only hits with min threshold
    query_target_fa_db_top_filt = path.join(output_dir, '%s_%s.tophit.minbitscore%s.mmsdb'
                                         % (query_prefix, target_prefix, bit_score_threshold))
    run_process(['mmseqs', 'filterdb', '--filter-column', '2', '--comparison-operator', 'ge', '--comparison-value',
                 str(bit_score_threshold), '--threads', str(threads), query_target_fa_db_top, query_target_fa_db_top_filt],
                verbose=verbose)
    # convert results to blast outformat 6
    forward_output_loc = path.join(output_dir, '%s_%s_hits.b6' % (query_prefix, target_prefix))
    run_process(['mmseqs', 'convertalis', query_db, target_db, query_target_db, forward_output_loc,
                 '--threads', str(threads)], verbose=verbose)
    return forward_output_loc


def get_reciprocal_best_hits(query_db, target_db, output_dir='.', query_prefix='query', target_prefix='target',
                             bit_score_threshold=60, rbh_bit_score_threshold=350, threads=10, verbose=False):
    """Take results from best hits and use for a reciprocal best hits search"""
    # TODO: Make it take query_target_db as a parameter
    # create subset for second search
    query_target_fa_db_top_filt = path.join(output_dir, '%s_%s.tophit.minbitscore%s.mmsdb'
                                         % (query_prefix, target_prefix, bit_score_threshold))  # I DON'T LIKE THIS
    query_target_fa_db_filt_top_swapped = path.join(output_dir, '%s_%s.minbitscore%s.tophit.swapped.mmsdb'
                                                 % (query_prefix, target_prefix, bit_score_threshold))
    # swap queries and targets in results database
    run_process(['mmseqs', 'swapdb', query_target_fa_db_top_filt, query_target_fa_db_filt_top_swapped, '--threads',
                 str(threads)], verbose=verbose)
    target_fa_db_filt = path.join(output_dir, '%s.filt.mmsdb' % target_prefix)
    # create a subdatabase of the target database with the best hits as well as the index of the target database
    run_process(['mmseqs', 'createsubdb', query_target_fa_db_filt_top_swapped, target_db, target_fa_db_filt], verbose=verbose)
    run_process(['mmseqs', 'createsubdb', query_target_fa_db_filt_top_swapped, '%s_h' % target_db,
                 '%s_h' % target_fa_db_filt], verbose=verbose)

    return get_best_hits(target_fa_db_filt, query_db, output_dir, target_prefix, query_prefix, rbh_bit_score_threshold,
                         threads, verbose)


def process_reciprocal_best_hits(forward_output_loc, reverse_output_loc, target_prefix='target'):
    """Process the forward and reverse best hits results to find reverse best hits
    Returns the query gene, target gene, if it was a reverse best hit, % identity, bit score and e-value
    """
    forward_hits = pd.read_csv(forward_output_loc, sep='\t', header=None, names=BOUTFMT6_COLUMNS)
    forward_hits = forward_hits.set_index('qId')
    reverse_hits = pd.read_csv(reverse_output_loc, sep='\t', header=None, names=BOUTFMT6_COLUMNS)
    reverse_hits = reverse_hits.set_index('qId')
    def check_hit(row:pd.Series):
        rbh = False
        if row.tId in reverse_hits.index:
            rbh = row.name == reverse_hits.loc[row.tId].tId
        return {'%s_hit' % target_prefix:      row.tId,
                '%s_RBH' % target_prefix:      rbh,
                '%s_identity' % target_prefix: row.seqIdentity,
                '%s_bitScore' % target_prefix: row.bitScore,
                '%s_eVal' % target_prefix:     row.eVal,
                'index':                       row.name
                }
    hits = forward_hits.apply(check_hit, axis=1, result_type='expand')
    # NOTE these lines may not be necessary
    hits.set_index('index', drop=True, inplace=True)
    hits.index.name = None
    return hits


def get_kegg_description(kegg_hits, header_dict):
    """Gets the KEGG IDs, and full KEGG hits from list of KEGG IDs for output in annotations"""
    gene_description = list()
    ko_list = list()
    for kegg_hit in kegg_hits.kegg_hit:
        header = header_dict[kegg_hit]
        gene_description.append(header)
        kos = re.findall(r'(K\d\d\d\d\d)', header)
        if len(kos) == 0:
            ko_list.append('')
        else:
            ko_list.append(','.join(kos))
    # TODO: change kegg_id to kegg_genes_id so that people get an error and not the wrong identifier
    new_df = pd.DataFrame([kegg_hits['kegg_hit'].values, ko_list, gene_description],
                          index=['kegg_genes_id', 'ko_id', 'kegg_hit'], columns=kegg_hits.index)
    return pd.concat([new_df.transpose(), kegg_hits.drop('kegg_hit', axis=1)], axis=1, sort=False)


def get_uniref_description(uniref_hits, header_dict):
    """Gets UniRef ID's, taxonomy and full string from list of UniRef IDs for output in annotations"""
    gene_description = list()
    uniref_list = list()
    gene_taxonomy = list()
    for uniref_hit in uniref_hits.uniref_hit:
        header = header_dict[uniref_hit]
        gene_description.append(header)
        uniref_list.append(header[header.find('RepID=') + 6:])
        gene_taxonomy.append(re.search(r'Tax=(.*?) (\S*?)=', header).group(1))
    new_df = pd.DataFrame([uniref_list, gene_description, gene_taxonomy],
                          index=['uniref_id', 'uniref_hit', 'uniref_taxonomy'],
                          columns=uniref_hits.index)
    return pd.concat([new_df.transpose(), uniref_hits.drop('uniref_hit', axis=1)], axis=1, sort=False)


def get_basic_description(hits, header_dict, db_name='viral', db_info_path=None):
    """Filter hits, and get descriptions
    (text before first space)"""
    hit_list = list()
    if db_info_path is not None:
        db_info = pd.read_csv(db_info_path, sep='\t', index_col=0)
        db_info.iloc[0]
        hits.iloc[0]
        hits[f"{db_name}_hits"]
        #hits = 
        hits[f"{db_name}_bitScore"]
        hits[f"{db_name}_bitScore"]
        hits = hits.merge(db_info, how='left',left_on=f"{db_name}_hit", 
                           right_index=True)
        hits.assign()
    description = list()
    for hit in hits['%s_hit' % db_name]:
        header = header_dict[hit]
        hit_list.append(hit)
        description.append(header)
    new_df = pd.DataFrame([hit_list, description],
                          index=['%s_id' % db_name, '%s_hit' % db_name],
                          columns=hits.index)
    return pd.concat([new_df.transpose(), hits.drop('%s_hit' % db_name, axis=1)], axis=1, sort=False)


def get_peptidase_description(peptidase_hits, header_dict):
    peptidase_list = list()
    peptidase_family = list()
    peptidase_descirption = list()
    for peptidase_hit in peptidase_hits.peptidase_hit:
        header = header_dict[peptidase_hit]
        peptidase_list.append(peptidase_hit)
        peptidase_family.append(re.search(r'#\w*.#', header).group()[1:-1])
        peptidase_descirption.append(header)
    new_df = pd.DataFrame([peptidase_list, peptidase_family, peptidase_descirption],
                          index=['peptidase_id', 'peptidase_family', 'peptidase_hit'], columns=peptidase_hits.index)
    return pd.concat([new_df.transpose(), peptidase_hits.drop('peptidase_hit', axis=1)], axis=1, sort=False)


def get_sig_row(row, evalue_lim:float=1e-15):
    """Check if hmm match is significant, based on dbCAN described parameters"""
    tstart, tend, tlen, evalue = row[['target_start', 'target_end', 'target_length', 'full_evalue']].values
    perc_cov = (tend - tstart)/tlen
    if perc_cov >= .35 and evalue <= evalue_lim:
        return True
    else:
        return False


# TODO: refactor following to methods to a shared run hmm step and individual get description steps
def parse_hmmsearch_domtblout(file):
    df_lines = list()
    for line in open(file):
        if not line.startswith('#'):
            line = line.split()
            line = line[:22] + [' '.join(line[22:])]
            df_lines.append(line)
    hmmsearch_frame = pd.DataFrame(df_lines, columns=HMMSCAN_ALL_COLUMNS)
    for i, column in enumerate(hmmsearch_frame.columns):
        hmmsearch_frame[column] = hmmsearch_frame[column].astype(HMMSCAN_COLUMN_TYPES[i])
    return hmmsearch_frame


def find_best_dbcan_hit(genome:str, group:pd.DataFrame):
    group['perc_cov'] = group.apply(
        lambda x: (x['target_end'] - x['target_start']) / x['target_length'],
        axis=1)
    group.sort_values('perc_cov', inplace=True)
    group.columns
    group.sort_values('full_evalue', inplace=True)
    return group.iloc[0]['target_id']


def run_hmmscan(genes_faa:str, db_loc:str, db_name:str, output_loc:str, formater:Callable,
                threads:int=2, verbose:bool=False):
    output = path.join(output_loc, f'{db_name}_results.unprocessed.b6')
    run_process(['hmmsearch', '--domtblout', output, '--cpu', str(threads), db_loc, genes_faa], verbose=verbose)
    #definition Parse hmmsearch output
    if not (path.isfile(output) and stat(output).st_size > 0):
        return pd.DataFrame()
    hits = parse_hmmsearch_domtblout(output)
    return formater(hits)


def get_gene_data(fasta_loc):
    """Take the prodigal gene headers and get the scaffold that it came from
    Based on idba_ud 'scaffold_#' scaffold names with gene name after
    """
    df_dict = dict()
    for seq in read_sequence(fasta_loc, format='fasta'):
        split_label = seq.metadata['id'].split('_')
        scaffold = '_'.join(split_label[:-1])
        gene_position = split_label[-1]
        start_position, end_position, strandedness = seq.metadata['description'].split('#')[1:4]
        df_dict[seq.metadata['id']] = [scaffold, int(gene_position), int(start_position), int(end_position),
                                       int(strandedness)]
    return pd.DataFrame.from_dict(df_dict, orient='index', columns=['scaffold', 'gene_position', 'start_position',
                                                                    'end_position', 'strandedness'])


def get_unannotated(fasta_loc, annotations):
    """Get the genes from the fasta which did not get any annotations"""
    return [seq.metadata['id'] for seq in read_sequence(fasta_loc, format='fasta')
            if seq.metadata['id'] not in annotations]


def get_minimum_bitscore(info_db):
    bit_score_threshold = min(info_db[['A_rank', 'B_rank']].min().values)
    return bit_score_threshold


def camper_blast_search(query_db, target_db, working_dir, info_db_path, start_time,
                          db_name, threads=10, verbose=False):
    """A convenience function to do a blast style reciprocal best hits search"""
    # Get kegg hits
    info_db = pd.read_csv(info_db_path, sep='\t', index_col=0)
    print('%s: Getting forward best hits from %s' % (str(datetime.now() - start_time), db_name))
    bit_score_threshold = get_minimum_bitscore(info_db)
    hits_path = get_best_hits(query_db, target_db, working_dir, 'gene', db_name, 
                         bit_score_threshold, threads, verbose=verbose)
    return camper_blast_search_formater(hits_path, db_name, info_db, start_time)

def bitScore_per_row(row):
    if row['score_type'] == 'domain':
        return row.domain_score
    elif row['score_type'] == 'full':
        return row.full_score
    elif row['score_type'] == '-':
        return None
    else:
        raise ValueError("The score_type must be 'domain', 'full', or 'j")

def rank_per_row(row):
    r_a = row['A_rank']
    r_b = row['B_rank']
    score = row['bitScore']
    if score is None:
        return None
    if float(score) >= float(r_a):
         return 'A'
    if pd.isnull(r_b):
        return None
    if float(score) >= float(r_b):
         return 'B'
    return None

#TODO decide if we need use_hmmer_thresholds:bool=False
def camper_hmmscan_formater(hits:pd.DataFrame,  db_name:str, 
                             hmm_info_path:str, top_hit:bool=True):
    if len(hits) < 1:
        return pd.DataFrame(columns = [
            f"{db_name}_id", f"{db_name}_rank", f"{db_name}_bitScore", 
            f"{db_name}_hits", f"{db_name}_search_type"])
    if hmm_info_path is None:
        hmm_info = None
        hits = hits[hits.apply(get_sig_row, axis=1)]
    else:
        hmm_info = pd.read_csv(hmm_info_path, sep='\t', index_col=0)
        hits = hits.merge(hmm_info, how='left',left_on="target_id", right_index=True)
        hits['bitScore'] = hits.apply(bitScore_per_row, axis=1)
        hits['score_rank'] = hits.apply(rank_per_row, axis=1)
        hits.dropna(subset=['score_rank'], inplace=True)
    if len(hits) == 0:
        # if nothing significant then return nothing, don't get descriptions
        return pd.DataFrame()
    if top_hit:
        # Get the best hits
        # TODO check we want top hit
        hits = hits.sort_values('full_evalue').drop_duplicates(subset=["query_id"])
    hits.set_index('query_id', inplace=True, drop=True)
    hits.rename_axis(None, inplace=True)
    if 'definition' in hits.columns:
        hits = hits[['target_id', 'score_rank', 'bitScore', 'definition']]
        hits.columns = [f"{db_name}_id", f"{db_name}_rank", 
                        f"{db_name}_bitScore", f"{db_name}_hits"]
    else:
        hits = hits[['target_id', 'score_rank', 'bitScore']]
        hits.columns = [f"{db_name}_id", f"{db_name}_rank", 
                        f"{db_name}_bitScore"]
    # Rename
    hits[f"{db_name}_search_type"] = 'hmm'
    return hits



def camper_blast_search_formater(hits_path, db_name, info_db, start_time):
    if stat(hits_path).st_size == 0:
        return pd.DataFrame()
    hits = pd.read_csv(hits_path, sep='\t', header=None, 
                       names=BOUTFMT6_COLUMNS, index_col='qId')
    hits = hits.merge(info_db, how='left',left_on="tId", right_index=True)
    rank_col = f"{db_name}_rank"
    hits[rank_col] = hits.apply(rank_per_row, axis=1)
    hits.dropna(subset=[rank_col], inplace=True)
    print('%s: Getting descriptions of hits from %s' % (str(datetime.now() - start_time), db_name))
    hits = hits[['tId', rank_col, 'bitScore', 'ID_for_distillate', 'definition']]
    hits.rename(columns={
            'tId': f"{db_name}_hits",
            'ID_for_distillate': f"{db_name}_id", 
            'bitScore': f"{db_name}_bitScore",  
            'definition': f"{db_name}_definition"
        },
        inplace=True)
    hits[f"{db_name}_search_type"] = 'blast'
    return hits


def count_motifs(gene_faa, motif='(C..CH)'):
    motif_count_dict = dict()
    for seq in read_sequence(gene_faa, format='fasta'):
        motif_count_dict[seq.metadata['id']] = len(list(seq.find_with_regex(motif)))
    return motif_count_dict


def strip_endings(text, suffixes: list):
    for suffix in suffixes:
        if text.endswith(suffix):
            text = text[:len(text) - len(suffix)]
    return text


def process_camper(output_dir, custom_fa_db_loc=(), custom_hmm_loc=(), 
                       custom_name=(), threads=1, verbose=False):
    mkdir(output_dir)
    if (len(custom_fa_db_loc) != len(custom_name)) or (len(custom_hmm_loc) != len(custom_name)):
        raise ValueError('Lengths of custom location list and custom name list must be the same.')
    custom_locs = dict()
    for name, fa_loc, hmm_loc in zip(custom_name, custom_fa_db_loc, custom_hmm_loc):
        new_fa_db_loc = path.join(output_dir, '%s.custom.mmsdb' % name)
        make_mmseqs_db(fa_loc, new_fa_db_loc, threads=threads, verbose=verbose)
        run_process(['hmmpress', '-f', hmm_loc], verbose=verbose)  # all are pressed just in case
        custom_locs[name] = {'fa': new_fa_db_loc, 'hmm': hmm_loc} 
    return custom_locs


def process_custom_hmm_cutoffs(custom_hmm_cutoffs_loc, custom_name, verbose=False):
    if custom_hmm_cutoffs_loc is None:
        return {}
    if len(custom_hmm_cutoffs_loc) < 0:
        return {}
    if custom_name is None:
        raise ValueError("You can't use the custom_hmm_cutoffs_loc argument without the custom_name and"
                         " custom_hmm_locs aguments specified.")
    if len(custom_hmm_cutoffs_loc) != len(custom_name):
        warnings.warn(f"Custom hmm cutoffs and descriptions were only provided to the first {len(custom_hmm_cutoffs_loc)}."
                      " The rest of the custom hmms will use standard cutoffs and have no descriptions.")
    return {custom_name[i]:j for i, j in enumerate(custom_hmm_cutoffs_loc)}

def process_custom_fa_db_cutoffs(custom_fa_db_cutoffs_loc, custom_name, verbose=False):
    if custom_fa_db_cutoffs_loc is None: 
        return {}
    if len(custom_fa_db_cutoffs_loc) < 0:
        return {}
    if custom_name is None:
        raise ValueError("You can't use the custom_fa_db_cutoffs_loc argument without the custom_name and"
                         " custom_fa_db_locs aguments specified.")
    if len(custom_fa_db_cutoffs_loc) != len(custom_name):
        warnings.warn(f"Custom fasta db cutoffs and descriptions were only provided to the first {len(custom_fa_db_cutoffs_loc)}."
                      " The rest of the custom fast dbs will use standard cutoffs and have no descriptions.")
    return {custom_name[i]:j for i, j in enumerate(custom_fa_db_cutoffs_loc)}


def annotate_orf(gene_faa:str, tmp_dir:str, start_time, 
                 custom_locs=(), 
                 custom_hmm_cutoffs_locs=(), 
                 custom_fa_db_cutoffs_locs=(), 
                 bit_score_threshold=60, rbh_bit_score_threshold=350,
                 threads=10, verbose=False):

    # Return a base dataframe so all genes are present
    yield pd.DataFrame( index=[seq.metadata['id'] for seq in read_sequence(gene_faa, format='fasta')])

    # Get kegg hits
    print('%s: Turning genes from prodigal to mmseqs2 db' % str(datetime.now() - start_time))
    query_db = path.join(tmp_dir, 'gene.mmsdb')
    make_mmseqs_db(gene_faa, query_db, create_index=True, threads=threads, verbose=verbose)

    for name, locs in custom_locs.items():
        print('%s: Getting hits from %s' % (str(datetime.now() - start_time), name))
        fasta = camper_blast_search(query_db=query_db, 
                                      target_db=locs['fa'], 
                                      working_dir=tmp_dir, 
                                      info_db_path=custom_fa_db_cutoffs_locs[name],
                                      start_time=start_time, 
                                      db_name=name, 
                                      threads=threads,
                                      verbose=verbose,)
        hmm = run_hmmscan(genes_faa=gene_faa,
                                           db_loc=locs['hmm'],
                                           db_name=name,
                                           threads=threads,
                                           output_loc=tmp_dir,
                                           formater=partial(
                                               camper_hmmscan_formater,
                                               db_name=name,
                                               hmm_info_path=custom_hmm_cutoffs_locs.get(name),
                                               top_hit=True
                                           ))
        full = pd.concat([fasta, hmm])
        yield (full
               .groupby(full.index)
               .apply(
                   lambda x: (x
                              .sort_values(f'{name}_search_type', ascending=True) # make sure hmm is first
                              .sort_values(f'{name}_bitScore', ascending=False)
                              .iloc[0])
                  )
               )



def check_fasta(input_faa):
    fasta_locs = glob(input_faa)
    if len(fasta_locs) == 0:
        raise ValueError('Given fasta locations returns no paths: %s' % input_faa)
    print('%s fastas found' % len(fasta_locs))
    return fasta_locs


def get_fa_db_name(fasta_loc):
    return path.splitext(path.basename(remove_suffix(fasta_loc, '.gz')))[0]

def merge_in_new_annotations(new_annotations:pd.DataFrame, past_annotations:pd.DataFrame):
    past_cols = set(past_annotations.columns)
    new_cols = set(new_annotations.columns)
    if new_cols.issubset(past_cols):
        raise Warning("You have passed an annotations file that containes"
                      " columns matching the new annotations. You most"
                      " likely are annotating with the same database again."
                      " The falowing columns will be replaced in the new"
                      " annotations file:\n%s" %
                      past_cols.intersection(new_cols))
    past_annotations = past_annotations[list(past_cols - new_cols)]
    while set(new_annotations.index) != set(past_annotations.index):
        if np.all([new_annotations.index.str.startswith('genes_')]): 
            new_annotations.index = new_annotations.index.str[6:]
            continue
        raise ValueError("The name in the genes.faa file dose not match the"
                         " annotations provided. Thus the old annotations can't"
                         " be merged to.")
    new_annotations = pd.merge(past_annotations, new_annotations, how="left", left_index=True, right_index=True)


def annotate_genes(input_faa, output_dir='.', bit_score_threshold=60, rbh_bit_score_threshold=350,
                   past_annotations_path:str=None, 
                   camper_fa_db_loc=(DEFAULT_CUSTOM_FA_DB_LOC), 
                   camper_fa_db_cutoffs_loc=(DEFAULT_CUSTOM_FA_DB_CUTOFFS_LOC,), 
                   camper_hmm_loc=(DEFAULT_CUSTOM_HMM_LOC),
                   camper_hmm_cutoffs_loc=(DEFAULT_CUSTOM_HMM_CUTOFFS_LOC), 
                   use_uniref=False, use_vogdb=False, 
                   rename_genes=True, keep_tmp_dir=True, threads=10, verbose=True):
    fasta_locs = glob(input_faa)
    if len(fasta_locs) == 0:
        raise ValueError('Given fasta locations returns no paths: %s' % input_faa)
    print('%s fastas found' % len(fasta_locs))
    # set up
    start_time = datetime.now()
    print('%s: Annotation started' % str(datetime.now()))

    if len(camper_fa_db_loc + camper_hmm_loc) < 1:
        raise ValueError("For some reason there are no database selected to"
                         " annotate against. This is most likely a result of"
                         " bad arguments.")

    mkdir(output_dir)
    tmp_dir = path.join(output_dir, 'working_dir')
    mkdir(tmp_dir)

    # setup camper databases to be searched
    camper_locs = process_camper(output_dir=path.join(tmp_dir, 'custom_dbs'), 
                                       custom_fa_db_loc=camper_fa_db_loc,
                                       custom_hmm_loc=camper_hmm_loc,
                                       custom_name=[CAMPER_NAME],  
                                       threads=threads,
                                       verbose=verbose)
    camper_fa_db_cutoffs_locs = process_custom_fa_db_cutoffs(camper_fa_db_cutoffs_loc, [CAMPER_NAME])
    camper_hmm_cutoffs_locs = process_custom_hmm_cutoffs(camper_hmm_cutoffs_loc, [CAMPER_NAME])
    print('%s: Retrieved database locations and descriptions' % (str(datetime.now() - start_time)))

    # annotate
    annotation_locs = list()
    faa_locs = list()
    # TODO IO is so slow!!! We need to use more memory and make less files
    for fasta_loc in fasta_locs:
        # set up
        fasta_name = get_fa_db_name(fasta_loc)
        fasta_dir = path.join(tmp_dir, fasta_name)
        mkdir(fasta_dir)

        # annotate
        annotations = pd.concat([
            orf for orf in
            annotate_orf(fasta_loc, fasta_dir,
                         start_time, camper_locs,
                         camper_hmm_cutoffs_locs, camper_fa_db_cutoffs_locs, 
                         bit_score_threshold, rbh_bit_score_threshold, 
                         threads, verbose)
                       ], axis=1)

        # add fasta name to frame and index, write file
        annotations.insert(0, 'fasta', fasta_name)
        if rename_genes:
            annotations.index = annotations.fasta + '_' + annotations.index
        annotation_loc = path.join(fasta_dir, 'annotations.tsv')
        annotations.to_csv(annotation_loc, sep='\t')
        annotation_locs.append(annotation_loc)

    # merge
    all_annotations = pd.concat([pd.read_csv(i, sep='\t', index_col=0) for i in annotation_locs], sort=False)
    all_annotations = all_annotations.sort_values('fasta')
    if past_annotations_path is not None:
        past_annotations = pd.read_csv(past_annotations_path, sep='\t', index_col=0)
        merge_in_new_annotations(new_annotations = all_annotations, 
                                 past_annotations = past_annotations)
    all_annotations.to_csv(path.join(output_dir, 'annotations.tsv'), sep='\t')
    if len(faa_locs) > 0:
        merge_files(faa_locs, path.join(output_dir, 'genes.faa'))

    # clean up
    if not keep_tmp_dir:
        rmtree(tmp_dir)

    print("%s: Completed annotations" % str(datetime.now() - start_time))


@click.command("annotate_genes")
@click.option('-i', '--input_faa', help="fasta file, optionally with wildcards to point to "
                                   "individual MAGs", required=True)
@click.option('-o', '--output_dir', help="output directory", required=True)
@click.option('-a', '--past_annotations_path',
              help="past_annotations to append new annotations to.")
# @click.option('--bit_score_threshold', type=int, default=60,
#          help='minimum bitScore of search to retain hits')
# @click.option('--rbh_bit_score_threshold', type=int, default=350,
#          help='minimum bitScore of reverse best hits to retain hits')
@click.option('--camper_fa_db_loc', default=[DEFAULT_CUSTOM_FA_DB_LOC], multiple=True,
         help="Location of CAMPER fastas to annotate against")
@click.option('--camper_fa_db_cutoffs_loc', default=[DEFAULT_CUSTOM_FA_DB_CUTOFFS_LOC], multiple=True,
              help="Location of file with camper fasta cutoffs and descriptions")
@click.option('--camper_hmm_loc',   default=[DEFAULT_CUSTOM_HMM_LOC], multiple=True,
              help="Location of camper hmms to annotate against")
@click.option('--camper_hmm_cutoffs_loc', default=[DEFAULT_CUSTOM_HMM_CUTOFFS_LOC], multiple=True,
              help="Location of file with CAMPER HMM cutoffs and descriptions")
# @click.option('--use_uniref', default=False,
#               help='Annotate these fastas against UniRef, drastically increases run time and '
#                    'memory requirements')
@click.option('--keep_tmp_dir', default=False)
@click.option('--threads', type=int, default=10, help='number of processors to use')
def annotate_genes_cmd(input_faa, output_dir='.', bit_score_threshold=60, rbh_bit_score_threshold=350,
                   past_annotations_path:str=None, 
                   camper_fa_db_loc=(DEFAULT_CUSTOM_FA_DB_LOC), 
                   camper_fa_db_cutoffs_loc=(DEFAULT_CUSTOM_FA_DB_CUTOFFS_LOC,), 
                   camper_hmm_loc=(DEFAULT_CUSTOM_HMM_LOC),
                   camper_hmm_cutoffs_loc=(DEFAULT_CUSTOM_HMM_CUTOFFS_LOC), 
                   use_uniref=False, use_vogdb=False, 
                   rename_genes=True, keep_tmp_dir=True, threads=10, verbose=True):
    annotate_genes(
        input_faa=input_faa,
        output_dir=output_dir,
        bit_score_threshold=bit_score_threshold,
        rbh_bit_score_threshold=rbh_bit_score_threshold,
        past_annotations_path=past_annotations_path,
        camper_fa_db_loc=camper_fa_db_loc,
        camper_fa_db_cutoffs_loc=camper_fa_db_cutoffs_loc,
        camper_hmm_loc=camper_hmm_loc,
        camper_hmm_cutoffs_loc=camper_hmm_cutoffs_loc,
        use_uniref=use_uniref,
        use_vogdb=use_vogdb,
        rename_genes=rename_genes,
        keep_tmp_dir=keep_tmp_dir,
        threads=threads,
        verbose=verbose)

if __name__ == "__main__":
    annotate_genes_cmd()

